import inspect
import logging
from abc import ABC, abstractmethod
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, ClassVar

import huggingface_hub
import llama_cpp
import openai

from evoprompt.cli import argument_parser
from evoprompt.opt_types import ModelUsage

logger = logging.getLogger(__name__)


current_directory = Path(__file__).resolve().parent


class LLMModel(ABC):
    models: ClassVar[dict[str, type["LLMModel"]]] = {}
    chat: bool

    def __init_subclass__(cls) -> None:
        if inspect.isabstract(cls):
            return
        cls.models[cls.__name__.lower()] = cls
        cls.register_arguments(argument_parser)

    @classmethod
    def get_model(cls, name: str, options: Namespace):
        if name not in cls.models:
            raise ValueError("Model %s does not exist", name)
        return cls.models[name](options)

    def __init__(self, options: Namespace):
        self.usage = ModelUsage()
        self.chat = options.chat

    @abstractmethod
    def __call__(
        self,
        system_message: str | None,
        prompt: str,
        *,
        prompt_appendix: str,
        prompt_prefix: str,
        prompt_suffix: str,
        chat: bool | None,
        stop: str,
        max_tokens: int,
        **kwargs: Any,
    ) -> Any:
        pass

    @classmethod
    @abstractmethod
    def register_arguments(cls, parser: ArgumentParser):
        pass


class Llama(LLMModel):

    def __init__(
        self,
        options: Namespace,
        n_gpu_layers: int = 60,
        n_threads: int = 8,
        n_ctx: int = 4096,
        **kwargs,
    ) -> None:

        super().__init__(options)

        # initialize model
        if options.llama_path is not None:
            # use local file
            self.model = llama_cpp.Llama(
                options.llama_path,
                chat_format=options.chat_format,
                chat_handler=options.chat_handler,
                verbose=options.verbose > 1 or options.llama_verbose,
                n_gpu_layers=n_gpu_layers,
                n_threads=n_threads,
                n_ctx=n_ctx,
                **kwargs,
            )
        else:
            # use pre-trained model from HF hub
            self.model = llama_cpp.Llama.from_pretrained(
                repo_id=options.llama_model,
                filename=options.llama_model_file,
                chat_format=options.chat_format,
                chat_handler=options.chat_handler,
                verbose=options.verbose > 1 or options.llama_verbose,
                n_gpu_layers=n_gpu_layers,
                n_threads=n_threads,
                n_ctx=n_ctx,
                **kwargs,
            )

    def __call__(
        self,
        system_message: str | None,
        prompt: str,
        *,
        prompt_appendix: str = "",
        prompt_prefix: str = "",
        prompt_suffix: str = "",
        chat: bool | None = None,
        stop: str = "</prompt>",
        max_tokens: int = 200,
        **kwargs: Any,
    ) -> tuple[str, ModelUsage]:
        if chat is None:
            chat = self.chat

        if chat:
            messages = [
                {
                    "role": "user",
                    "content": prompt_prefix + prompt + prompt_suffix + prompt_appendix,
                }
            ]
            if system_message:
                messages.insert(
                    0,
                    {
                        "role": "system",
                        "content": system_message,
                    },
                )
            response = self.model.create_chat_completion(
                messages=messages,
                stop=stop,
                max_tokens=max_tokens,
                **kwargs,
            )
            response_text = response["choices"][0]["message"]["content"]
        else:
            response = self.model.create_completion(
                prompt=(system_message if system_message else "")
                + prompt_prefix
                + prompt
                + prompt_suffix
                + prompt_appendix,
                stop=stop,
                max_tokens=max_tokens,
                **kwargs,
            )
            response_text = response["choices"][0]["text"]
        # input(f"Response: {response_text}")
        usage = ModelUsage(**response["usage"])
        return response_text, usage

    @classmethod
    def register_arguments(cls, parser: ArgumentParser):
        group = parser.add_argument_group(f"{cls.__name__} model arguments")
        group.add_argument(
            "--llama-path",
            type=str,
            help="Specify path to local Llama model, takes precedence over --llama-model",
        ),
        group.add_argument(
            "--llama-model",
            type=str,
            default="TheBloke/Llama-2-13B-chat-GGUF",
            help="A pre-trained model from HF hub",
        ),
        group.add_argument(
            "--llama-model-file",
            type=str,
            # TODO provide some help for selecting model files, and point user to set this argument if needed
            default="llama-2-13b-chat.Q5_K_M.gguf",
            help="Specify the model file in case of a pre-trained model from HF hub, e.g., a specific quantized version",
        ),
        group.add_argument(
            "--chat-format",
            type=str,
            help="Override chat format (--chat-handler takes precedence if specified)",
        )
        group.add_argument(
            "--chat-handler",
            type=str,
            help="Override chat handler (takes precedence over --chat-format)",
        )
        group.add_argument(
            "--llama-verbose",
            "-v",
            action="store_true",
            help="Increase verbosity of Llama model",
        )


class OpenAI(LLMModel):
    """Queries an OpenAI model using its API."""

    def __init__(
        self,
        options: Namespace,
        verbose: bool = False,
        **kwargs,
    ) -> None:
        self.model_name = options.openai_model
        super().__init__(options)

        # initialize client for API calls
        self.openai_client = openai.OpenAI(**kwargs)

    def __call__(
        self,
        system_message: str | None,
        prompt: str,
        *,
        prompt_appendix: str = "",
        prompt_prefix: str = "",
        prompt_suffix: str = "",
        chat: bool | None = None,
        stop: str = "</prompt>",
        max_tokens: int = 200,
        **kwargs: Any,
    ) -> tuple[str, ModelUsage]:
        if chat is None:
            chat = self.chat

        if chat:
            messages = [
                {
                    "role": "user",
                    "content": prompt_prefix + prompt + prompt_suffix + prompt_appendix,
                }
            ]
            if system_message:
                messages.insert(
                    0,
                    {
                        "role": "system",
                        "content": system_message,
                    },
                )
            response = self.openai_client.chat.completions.create(
                model=self.model_name,
                messages=messages,
                stop=stop,
                max_tokens=max_tokens,
                **kwargs,
            )
            usage = ModelUsage(**response.usage.__dict__)
            return response.choices[0].message.content, usage
        else:
            response = self.openai_client.completions.create(
                model=self.model,
                prompt=(system_message if system_message else "")
                + prompt_prefix
                + prompt
                + prompt_suffix
                + prompt_appendix,
                stop=stop,
                max_tokens=max_tokens,
                **kwargs,
            )
            usage = ModelUsage(**response.usage.__dict__)
            return response.choices[0].text, usage

    @classmethod
    def register_arguments(cls, parser: ArgumentParser):
        group = parser.add_argument_group("OpenAI model arguments")
        group.add_argument("--openai-model", "-m", type=str, default="gpt-3.5-turbo")


argument_group = argument_parser.add_argument_group("Model arguments")
argument_group.add_argument(
    "--evolution-engine",
    "-e",
    type=str,
    choices=LLMModel.models.keys(),
    default="llama",
)
argument_group.add_argument("--chat", "-c", action="store_true")
