import logging
from abc import ABCMeta, abstractmethod
from collections import deque
from dataclasses import KW_ONLY, dataclass
from statistics import mean
from typing import Iterable, Literal, Union

from datasets import Dataset, load_dataset
from llama_cpp import LlamaGrammar
from tqdm import tqdm

from evoprompt.models import LLMModel
from evoprompt.opt_types import ModelUsage
from evoprompt.utils import log_calls

logger = logging.getLogger(__name__)


SYSTEM_MESSAGE = """
You are given an instruction that describes a task, paired with an input that provides further context. Write a response that appropriately completes the request.
"""

DatasetDatum = dict


ParentHistories = list[list[float]]


class EarlyStoppingMonitor:

    @abstractmethod
    def update(self, score: float) -> bool:
        raise NotImplementedError


@dataclass
class MomentBasedStopping(EarlyStoppingMonitor):
    """
    Watch the first derivative (moment) of the metric to determine when to stop.
    """

    _: KW_ONLY
    patience: int = 10
    start_after: int = 20
    min_moment_magnitude: float = 0.001

    def __post_init__(self):
        self.moment_magnitudes = deque(maxlen=self.patience)
        self.last_score = 0.0
        self.num_calls = 0

    def update(self, score: float) -> bool:
        # caclulate the current moment (dx/dt)
        self.num_calls += 1
        if self.num_calls < self.start_after:
            return False

        self.moment_magnitudes.append(abs(score - self.last_score))
        self.last_score = score
        if len(self.moment_magnitudes) < self.patience:
            return False

        if mean(self.moment_magnitudes) < self.min_moment_magnitude:
            return True

        return False


@dataclass
class ForcedImprovementBasedStopping(EarlyStoppingMonitor):
    """
    let the model run for a fixed number of steps and then stop if the model
    has not improved by a certain amount.
    """

    _: KW_ONLY
    patience: int = 10
    minimum_score_improvement: float = 0.001

    def __post_init__(self):
        self.scores = []

    def update(self, score: float) -> bool:
        self.scores.append(score)

        if len(self.scores) >= self.patience:
            return mean(self.scores) < self.minimum_score_improvement

        return False


@dataclass
class ParentBaselineBasedStopping(EarlyStoppingMonitor):

    parent_histories: ParentHistories
    _: KW_ONLY
    patience: int = 10
    start_after: int = 20
    min_improvement: float = 0.001

    def __post_init__(self):
        self.num_calls = 0
        self.improvement_memory = deque(maxlen=self.patience)

    def update(self, score: float) -> bool:
        self.num_calls += 1
        if self.num_calls < self.start_after:
            return False

        parent_values = [  # get the metric value of the parents at the current step
            (
                parent_history[self.num_calls - 1]
                if len(parent_history) >= self.num_calls
                else parent_history[-1]  # extend with last value
            )
            for parent_history in self.parent_histories
        ]
        self.improvement_memory.append(
            score - max(parent_values)  # compare with the best parent
        )

        if len(self.improvement_memory) < self.patience:
            return False

        if max(self.improvement_memory) < self.min_improvement:
            # if the highest improvement is less than the minimum improvement, we stop
            return True

        return False


EvaluationStrategyKey = Literal[
    "simple", "early-stopping", "shortest-first", "hardest-first"
]


def get_evaluation_strategy(evaluation_strategy_key: EvaluationStrategyKey):
    match evaluation_strategy_key:
        case "simple":
            return SimpleStrategy
        case "early-stopping":
            return EarlyStoppingStrategy
        case "shortest-first":
            return ShortestFirstStrategy
        case "hardest-first":
            return HardestFirstStrategy


class EvaluationStrategy:
    def __init__(self, task):
        self.task = task

    @abstractmethod
    def get_dataset_iterator(
        self, dataset: Dataset, parent_histories: ParentHistories | None
    ) -> Iterable[DatasetDatum]:
        raise NotImplementedError

    @abstractmethod
    def should_stop(self, score: float) -> bool:
        raise NotImplementedError


class SimpleStrategy(EvaluationStrategy):

    def get_dataset_iterator(
        self, dataset: Dataset, parent_histories: ParentHistories | None
    ):
        return tqdm(dataset, desc="evaluating prompt", leave=False)

    def should_stop(self, score: float) -> bool:
        return False


class EarlyStoppingStrategy(SimpleStrategy):
    def get_dataset_iterator(self, dataset, parent_histories):
        early_stopping_params = {
            "patience": max(len(dataset) // 20, 5),
            "start_after": max(len(dataset) // 5, 5),
        }
        if parent_histories is not None:
            self.early_stopping = ParentBaselineBasedStopping(
                parent_histories, **early_stopping_params
            )
        else:
            self.early_stopping = MomentBasedStopping(**early_stopping_params)

        return super().get_dataset_iterator(dataset, parent_histories)

    def should_stop(self, score: float):
        return self.early_stopping.update(score)


class ShortestFirstStrategy(EarlyStoppingStrategy):
    def get_dataset_iterator(
        self, dataset: Dataset, parent_histories: ParentHistories | None
    ):
        sorted_dataset = sorted(
            dataset, key=lambda x: len(self.task._get_prompt_text_for_datum(x))
        )
        return super().get_dataset_iterator(sorted_dataset, parent_histories)


class HardestFirstStrategy(ShortestFirstStrategy):
    def get_dataset_iterator(
        self, dataset: Dataset, parent_histories: ParentHistories | None
    ):
        if parent_histories is None:
            # if we don't have parent histories, we can't sort by difficulty
            # so we fall back to the shortest first strategy
            return super().get_dataset_iterator(dataset, parent_histories)

        # get the metric changes for each parent, a positive change means
        # improvement and therefore a correct prediction
        metric_changes = [
            [0] + [p1 - p0 for p0, p1 in zip(history[0:], history[1:])]
            for history in parent_histories
        ]

        # calculate the average metric change for each sample
        average_metric_changes = [
            mean(parent_scores) for parent_scores in zip(*metric_changes)
        ]

        # we use a forced improvement based stopping criterion to stop the
        # evaluation when the model does just as bad as the parents on the hardest samples
        num_hard_samples = len(list(filter(lambda x: x <= 0, average_metric_changes)))
        self.early_stopping = ForcedImprovementBasedStopping(
            patience=num_hard_samples, minimum_score_improvement=0.001
        )

        # sort the dataset by the average metric change so that the hardest
        # (most negative change) samples come first
        sorted_dataset = [
            datum
            for datum, _change in sorted(
                zip(dataset, average_metric_changes),
                key=lambda x: x[1],
            )
        ]

        # pass the EarlyStoppingStrategy in the MRO to not override the early_stopping attribute
        return super(EarlyStoppingStrategy, self).get_dataset_iterator(
            sorted_dataset, parent_histories
        )

    def should_stop(self, score: float):
        return self.early_stopping.update(score)


class Task(metaclass=ABCMeta):
    shorthand: str
    validation_dataset: Dataset
    test_dataset: Dataset

    def __init__(
        self,
        model: Union[LLMModel],
        validation_dataset: str | None = None,
        test_dataset: str | None = None,
        *,
        use_grammar: bool,
        evaluation_strategy: EvaluationStrategyKey,
        validation_split: str | None = None,
        test_split: str | None = None,
        debug: bool = False,
        **kwargs,
    ) -> None:
        self.model = model
        self.debug = debug
        # whether we use the grammar to constrain the model output or not
        self.use_grammar = use_grammar
        self.evaluation_strategy = get_evaluation_strategy(evaluation_strategy)(self)

        self.validation_dataset = self.load_validation_set(
            validation_dataset, validation_split
        )
        if self.debug and len(self.validation_dataset) > 10:
            self.validation_dataset = self.validation_dataset.shuffle(42).select(
                range(10)
            )
        else:
            # NOTE currently we select a subset for validation
            if len(self.validation_dataset) > 200:
                self.validation_dataset = self.validation_dataset.shuffle(42).select(
                    range(200)
                )

        self.test_dataset = self.load_test_set(test_dataset, test_split)
        if self.debug and len(self.test_dataset) > 5:
            self.test_dataset = self.test_dataset.select(range(5))

    def load_validation_set(
        self, validation_dataset: str, validation_split: str | None
    ):
        return load_dataset(validation_dataset, split=validation_split)

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(test_dataset, split=test_split)

    def predict(self, prompt: str, datum: DatasetDatum) -> tuple[str, ModelUsage]:
        # run model for inference using grammar to constrain output
        # TODO grammar also depends on prompt and vice-versa -> what are good labels?
        response, usage = self.model(
            system_message=SYSTEM_MESSAGE,
            prompt=prompt,
            prompt_appendix=self._get_prompt_text_for_datum(datum),
            grammar=self._get_grammar(datum) if self.use_grammar else None,
        )

        if not self.use_grammar:
            # we postprocess the model output to return as answer
            response = response.strip()

        return response, usage

    @abstractmethod
    def _get_grammer(self, datum: DatasetDatum) -> LlamaGrammar:
        pass

    @abstractmethod
    def _evaluate_sample(
        self, prompt: str, datum: DatasetDatum
    ) -> tuple[str, ModelUsage]:
        pass

    @abstractmethod
    def _get_prompt_text_for_datum(self, datum: DatasetDatum) -> str:
        pass

    @abstractmethod
    def _aggregate_result(self, results: list) -> float:
        pass

    def evaluate(
        self,
        prompt: str,
        dataset: Dataset,
        parent_histories: list[list[float]] | None = None,
        no_early_stopping: bool = False,
    ) -> tuple[float, ModelUsage, list[float]]:
        dataset_iterator = self.evaluation_strategy.get_dataset_iterator(
            dataset, parent_histories
        )

        logger.info(
            f"using evaluation strategy: {self.evaluation_strategy.__class__.__name__}"
        )
        if hasattr(self.evaluation_strategy, "early_stopping"):
            logger.info(
                f"using early stopping: {self.evaluation_strategy.early_stopping}",
            )

        results: list = []
        evaluation_usage = ModelUsage()
        evaluation_history = []

        for datum in dataset_iterator:
            result, usage = self._evaluate_sample(prompt, datum)
            results.append(result)
            current_metric = self._aggregate_result(results)
            dataset_iterator.set_postfix({self.metric_name: f"{current_metric:.2f}"})
            evaluation_usage += usage
            evaluation_history.append(current_metric)
            if not no_early_stopping and self.evaluation_strategy.should_stop(
                current_metric
            ):
                logger.info(
                    f"Early stopping after {len(results)} samples with {self.metric_name} of {current_metric:.2f}"
                )
                break

        return self._aggregate_result(results), evaluation_usage, evaluation_history

    @log_calls("Evaluating validation dataset")
    def evaluate_validation(
        self, prompt: str, parent_histories: list[list[float]] | None = None
    ):
        return self.evaluate(prompt, self.validation_dataset, parent_histories)

    @log_calls("Evaluating test dataset")
    def evaluate_test(self, prompt: str):
        return self.evaluate(prompt, self.test_dataset, no_early_stopping=True)

    @property
    @abstractmethod
    def metric_name(self) -> str:
        pass

    @property
    @abstractmethod
    def base_prompt(self) -> str:
        pass
