from functools import lru_cache
from typing import Mapping

from datasets import load_dataset
from evoprompt.models import LLMModel
from evoprompt.task import TextClassification
from evoprompt.task.task import DatasetDatum


class AGNews(TextClassification):
    shorthand = "agn"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(
            *args,
            validation_dataset="fancyzhx/ag_news",
            test_dataset="fancyzhx/ag_news",
            validation_split="train",
            test_split="test",
            **kwargs
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["text"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["world", "sports", "business", "tech"]
        return dict(zip(classes, range(len(classes))))

    @property
    def base_prompt(self):
        #  from the paper: RLPROMPT: Optimizing Discrete Text Prompts with Reinforcement Learning
        return 'In this task, you are given a news article. Your task is to classify the article to one out of the four topics "World", "Sports", "Business", "Tech" if the article"s main topic is relevant to the world, sports, business, and technology, correspondingly. If you are not sure about the topic, choose the closest option.'


class TREC(TextClassification):
    shorthand = "trec"

    def load_validation_set(
        self, validation_dataset: str | None, validation_split: str | None
    ):
        return load_dataset(
            "csv",
            data_files={"validation": "evoprompt/data/trec/train.csv"},
            split="validation",
            column_names=["label", "question"],
        )

    def load_test_set(self, test_dataset: str, test_split: str | None):
        return load_dataset(
            "csv",
            data_files={"test": "evoprompt/data/trec/test.csv"},
            split="test",
            column_names=["label", "question"],
        )

    def _get_text_for_datum(self, datum: DatasetDatum) -> str:
        return datum["question"]

    def _get_gold_label_for_datum(self, datum: DatasetDatum) -> str:
        return datum["label"]

    @lru_cache
    def _get_label_mapping(self) -> Mapping:
        classes = ["description", "entity", "expression", "human", "location", "number"]
        return dict(zip(classes, range(len(classes))))

    @property
    def base_prompt(self):
        #  from the paper: RLPROMPT: Optimizing Discrete Text Prompts with Reinforcement Learning
        return 'You are given a question. You need to detect which category better describes the question. Answer with "Description", "Entity", "Expression", "Human", "Location", and "Number".'
